
CONTACTS_MAP = {
    'allianceContactList': 'alliance',
    'corporateContactList': 'corp',
    'contactList': 'personal',
}

LABEL_MAP = {
    'contactLabels': 'personal',
    'corporateContactLabels': 'corp',
    'allianceContactLabels': 'alliance',
}


def parse_contact_list(api_result):
    result = {'labels': {}}
    for rowset in api_result.findall('rowset'):
        setname = rowset.get('name')

        if setname in CONTACTS_MAP:
            contact_list = result[CONTACTS_MAP[rowset.get('name')]] = {}
            for row in rowset.findall('row'):
                in_watchlist = (row.get('inWatchlist') == 'True'
                                if 'inWatchlist' in row.attrib
                                else None)
                contact_id = int(row.get('contactID'))
                contact_list[contact_id] = {
                    'id': contact_id,
                    'name': row.get('contactName'),
                    'standing': float(row.get('standing')),
                    'in_watchlist': in_watchlist,
                    'label_mask': int(row.get('labelMask') or 0),
                    'labels': {},
                }
        elif setname in LABEL_MAP:
            label_list = result['labels'][LABEL_MAP[setname]] = {}
            for row in rowset.findall('row'):
                label_id = int(row.get('labelID'))
                label_list[label_id] = {
                    'id': label_id,
                    'name': row.get('name'),
                }
    for grouping in ('personal', 'corp', 'alliance'):
        group = result.get(grouping)
        if not group:
            continue
        labels = result['labels'][grouping]
        for contact_id in group:
            contact = group[contact_id]
            labelMask = contact['label_mask']
            if labelMask:
                for label_id in labels:
                    if labelMask & label_id:
                        contact['labels'][label_id] = labels[label_id]
            del contact['label_mask']

    return result



def parse_assets(api_result):
    def handle_rowset(rowset, parent_location):
        results = []
        for row in rowset.findall('row'):
            item = {'id': int(row.attrib['itemID']),
                    'item_type_id': int(row.attrib['typeID']),
                    'location_id': int(row.attrib.get('locationID', parent_location)),
                    'location_flag': int(row.attrib['flag']),
                    'quantity': int(row.attrib['quantity']),
                    'packaged': row.attrib['singleton'] == '0',
            }
            raw_quantity = row.attrib.get('rawQuantity')
            if raw_quantity is not None:
                item['raw_quantity'] = int(raw_quantity)
            contents = row.find('rowset')
            if contents is not None:
                item['contents'] = handle_rowset(contents, item['location_id'])
            results.append(item)
        return results

    result_list = handle_rowset(api_result.find('rowset'), None)
    # For convenience, key the result by top-level location ID.
    result_dict = {}
    for item in result_list:
        location = item['location_id']
        result_dict.setdefault(location, {})
        result_dict[location]['location_id'] = location
        result_dict[location].setdefault('contents', [])
        result_dict[location]['contents'].append(item)
    return result_dict



from evelink import api
from evelink import constants

def parse_industry_jobs(api_result):
        rowset = api_result.find('rowset')
        result = {}

        if rowset is None:
            return

        for row in rowset.findall('row'):
            # shortcut to make the following block less painful
            a = row.attrib
            jobID = int(a['jobID'])
            completed = a['completedCharacterID'] != '0'
            result[jobID] = {
                'activity_id': int(a['activityID']),
                'blueprint': {
                    'id': int(a['blueprintID']),
                    'location_id': int(a['blueprintLocationID']),
                    'type': {
                        'id': int(a['blueprintTypeID']),
                        'name': a['blueprintTypeName'],
                    },
                },
                'completed': completed,
                'complete_ts': api.parse_ts(a['completedDate']),
                'completor_id': int(a['completedCharacterID']),
                'cost': float(a['cost']),
                'end_ts': api.parse_ts(a['endDate']),
                'facility_id': int(a['facilityID']),
                'installer': {
                    'id': int(a['installerID']),
                    'name': a['installerName'],
                },
                'product': {
                    'type_id': int(a['productTypeID']),
                    'location_id': int(a['outputLocationID']),
                    'name': a['productTypeName'],
                    'probability': float(a['probability']),
                },
                'runs': int(a['runs']),
                'licensed_runs': int(a['licensedRuns']),
                'pause_ts': api.parse_ts(a['pauseDate']),
                'system': {
                    'id': int(a['solarSystemID']),
                    'name': a['solarSystemName'],
                },
                'station_id': int(a['stationID']),
                'begin_ts': api.parse_ts(a['startDate']),
                'status': int(a['status']),
                'team_id': int(a['teamID']),
                'duration': int(a['timeInSeconds']),
            }

        return result



from evelink import account
from evelink.appengine.api import auto_async

@auto_async
class Account(account.Account):
    __doc__ = account.Account.__doc__

    def __init__(self, api):
        self.api = api



ACCOUNT = 'account'
CHARACTER = 'char'
CORPORATION = 'corp'

BLUEPRINT_ORIGINAL = -1
BLUEPRINT_COPY = -2

_role_type_bases = {
    'global': '',
    'at_hq': 'AtHQ',
    'at_base': 'AtBase',
    'at_other': 'AtOther',
}

class Char(object):
    corp_roles = dict((k, 'corporationRoles' + v) for k,v in _role_type_bases.items())

class Corp(object):
    role_types = dict((k, 'roles' + v) for k,v in _role_type_bases.items())
    grantable_types = dict((k, 'grantableRoles' + v) for k,v in _role_type_bases.items())

    pos_states = ('unanchored', 'anchored', 'onlining', 'reinforced', 'online')

    pos_permission_entities = (
            'Starbase Config',
            'Starbase Fuel Tech',
            'Corporation Members',
            'Alliance Members',
        )

class Industry(object):
    job_status = ('failed', 'delivered', 'aborted', 'gm-aborted', 'inflight-unanchored', 'destroyed')

class Market(object):
    order_status = ('active', 'closed', 'expired', 'cancelled', 'pending', 'deleted')

class APIKey(object):
    key_types = {
        # This maps from EVE API values (keys) to our local constants (values)
        'Account': ACCOUNT,
        'Character': CHARACTER,
        'Corporation': CORPORATION,
    }



"""EVELink - Python bindings for the EVE API."""

import logging

from evelink import account
from evelink import api
from evelink import char
from evelink import constants
from evelink import corp
from evelink import eve
from evelink import map
from evelink import server

__version__ = "0.7.2"

# Implement NullHandler because it was only added in Python 2.7+.
class NullHandler(logging.Handler):
    def emit(self, record):
        pass

# Create a logger, but by default, have it do nothing
_log = logging.getLogger('evelink')
_log.addHandler(NullHandler())

# Update the version number used in the user-agent
api._user_agent = 'evelink v%s' % __version__

__all__ = [
  "account",
  "api",
  "char",
  "constants",
  "corp",
  "eve",
  "map",
  "parsing",
  "server",
]



from evelink.appengine.api import AppEngineAPI
from evelink.appengine.api import AppEngineCache
from evelink.appengine.api import AppEngineDatastoreCache
from evelink.appengine.api import EveLinkCache
from evelink.appengine import account
from evelink.appengine import char
from evelink.appengine import corp
from evelink.appengine import eve
from evelink.appengine import map
from evelink.appengine import server

__all__ = [
  "AppEngineAPI",
  "AppEngineCache",
  "AppEngineDatastoreCache",
  "EveLinkCache",
  "account",
  "char",
  "corp",
  "eve",
  "map",
  "server",
]



import functools
import inspect
import time
from urllib import urlencode
from xml.etree import ElementTree

from google.appengine.api import memcache
from google.appengine.api import urlfetch
from google.appengine.ext import ndb

from evelink import api


class UrlFetchError(Exception): pass


class AppEngineAPI(api.API):
    """Subclass of api.API that is compatible with Google Appengine."""

    def __init__(self, base_url="api.eveonline.com", cache=None, api_key=None):
        cache = cache or AppEngineCache()
        super(AppEngineAPI, self).__init__(base_url=base_url,
                cache=cache, api_key=api_key)

    @ndb.tasklet
    def get_async(self, path, params=None):
        """Asynchronous request a specific path from the EVE API.

        TODO: refactor evelink.api.API.get
        """

        params = params or {}
        params = dict((k, api._clean(v)) for k,v in params.items())

        if self.api_key:
            params['keyID'] = self.api_key[0]
            params['vCode'] = self.api_key[1]

        key = self._cache_key(path, params)
        response = yield self.cache.get_async(key)
        cached = response is not None

        if not cached:
            # no cached response body found, call the API for one.
            params = urlencode(params)
            full_path = "https://%s/%s.xml.aspx" % (self.base_url, path)
            response, robj = yield self.send_request_async(full_path, params)

        try:
            tree = ElementTree.fromstring(response)
        except ElementTree.ParseError as e:
            self.maybe_raise_http_error(robj)
            raise e

        current_time = api.get_ts_value(tree, 'currentTime')
        expires_time = api.get_ts_value(tree, 'cachedUntil')
        self._set_last_timestamps(current_time, expires_time)

        if not cached:
            yield self.cache.put_async(key, response, expires_time - current_time)

        error = tree.find('error')
        if error is not None:
            code = error.attrib['code']
            message = error.text.strip()
            exc = api.APIError(code, message, current_time, expires_time)
            raise exc

        result = tree.find('result')
        raise ndb.Return(api.APIResult(result, current_time, expires_time))

    def maybe_raise_http_error(self, robj):
        if robj.status_code != 200:
            raise UrlFetchError(
                "HTTP error {0}".format(robj.status_code),
                robj.status_code)

    def send_request(self, url, params):
        """Send a request via the urlfetch API.

        url:
            The url to fetch
        params:
            URL encoded parameters to send. If set, will use a form POST,
            otherwise a GET.
        """
        params = urlencode(params)
        return self.send_request_async(url, params).get_result()

    @ndb.tasklet
    def send_request_async(self, url, params):
        ctx = ndb.get_context()
        headers = {'User-agent': self.user_agent}
        if params:
            headers['Content-Type'] = 'application/x-www-form-urlencoded'
        result = yield ctx.urlfetch(
            url=url,
            payload=params,
            method=urlfetch.POST if params else urlfetch.GET,
            headers=headers,
            deadline=api.http_request_timeout,
        )
        raise ndb.Return((result.content, result))


class AppEngineCache(api.APICache):
    """Memcache backed APICache implementation."""

    def get(self, key):
        return memcache.get(key)

    @ndb.tasklet
    def get_async(self, key):
        """Dummy async method.

        Memcache doesn't have an async get method.
        """
        raise ndb.Return(self.get(key))


    def put(self, key, value, duration):
        if duration < 0:
            duration = time.time() + duration
        memcache.set(key, value, time=duration)

    @ndb.tasklet
    def put_async(self, key,  value, duration):
        """Dummy async method (see get_async)."""
        self.put(key, value, duration)


class EveLinkCache(ndb.Model):
    value = ndb.PickleProperty()
    expiration = ndb.IntegerProperty()


class AppEngineDatastoreCache(api.APICache):
    """An implementation of APICache using the AppEngine datastore."""

    def __init__(self):
        super(AppEngineDatastoreCache, self).__init__()

    def get(self, cache_key):
        return self.get_async(cache_key).get_result()

    @ndb.tasklet
    def get_async(self, cache_key):
        db_key = ndb.Key(EveLinkCache, cache_key)
        result = yield db_key.get_async()

        if not result:
            raise ndb.Return(None)

        if result.expiration < time.time():
            yield db_key.delete_async()
            raise ndb.Return(None)

        raise ndb.Return(result.value)

    def put(self, cache_key, value, duration):
        self.put_async(cache_key, value, duration).get_result()

    @ndb.tasklet
    def put_async(self, cache_key, value, duration):
        expiration = int(time.time() + duration)
        cache = EveLinkCache(id=cache_key, value=value, expiration=expiration)
        yield cache.put_async()


def auto_gae_api(func):
    """A decorator to automatically provide an AppEngineAPI instance."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        defaultargs, defaultkwargs = api.get_args_and_defaults(func)
        mapped_args = api.map_func_args(args, kwargs, defaultargs, defaultkwargs)
        if mapped_args.get('api') is None:
            kwargs['api'] = AppEngineAPI()
        return func(*args, **kwargs)
    return wrapper


def _make_async(method):
    def _async(self, *args, **kw):
        # method specs
        path = method._request_specs['path']
        args_names = method._request_specs['args']
        defaults = method._request_specs['defaults']
        prop_to_param = method._request_specs['prop_to_param']
        map_params = method._request_specs['map_params']

        # build parameter map
        args_map = api.map_func_args(args, kw, args_names, defaults)
        for attr_name in prop_to_param:
            args_map[attr_name] = getattr(self, attr_name, None)

        # fix params name and remove params with None values
        params = api.translate_args(args_map, map_params)
        params =  dict((k, v,) for k, v in params.items() if v is not None)

        kw['api_result'] = yield self.api.get_async(path, params=params)
        raise ndb.Return(method(self, *args, **kw))
    return ndb.tasklet(_async)


def auto_async(cls):
    """Class decoration which add a async version of any method with a
    a '_request_specs' attribute (metadata added by api.auto_add).
    """
    for method_name, method in inspect.getmembers(cls, inspect.ismethod):
        if not hasattr(method, '_request_specs'):
            continue

        async_method = _make_async(method)
        async_method.__doc__ = """Asynchronous version of %s.""" % method_name
        async_method.__name__ = '%s_async' % method_name
        setattr(cls, async_method.__name__, async_method)

    return cls



from evelink import api

class Map(object):
    """Wrapper around /map/ of the EVE API."""

    @api.auto_api
    def __init__(self, api=None):
        self.api = api

    @api.auto_call('map/Jumps')
    def jumps_by_system(self, api_result=None):
        """Get jump counts for systems in the last hour.

        Returns a tuple of ({system:jumps...}, timestamp).

        NOTE: Systems with 0 jumps in the last hour are not included!
        """

        rowset = api_result.result.find('rowset')
        results = {}
        for row in rowset.findall('row'):
            system = int(row.attrib['solarSystemID'])
            jumps = int(row.attrib['shipJumps'])
            results[system] = jumps

        data_time = api.parse_ts(api_result.result.find('dataTime').text)

        return api.APIResult((results, data_time), api_result.timestamp, api_result.expires)

    @api.auto_call('map/Kills')
    def kills_by_system(self, api_result=None):
        """Get kill counts for systems in the last hour.

        Returns a tuple of ({system:{killdata}, timestamp).

        Each {killdata} is {'faction':count, 'ship':count, 'pod':count}.
        """
        
        rowset = api_result.result.find('rowset')
        results = {}
        for row in rowset.findall('row'):
            system = int(row.attrib['solarSystemID'])
            faction_kills = int(row.attrib['factionKills'])
            ship_kills = int(row.attrib['shipKills'])
            pod_kills = int(row.attrib['podKills'])

            results[system] = {
                'id': system,
                'faction': faction_kills,
                'ship': ship_kills,
                'pod': pod_kills,
            }

        data_time = api.parse_ts(api_result.result.find('dataTime').text)

        return api.APIResult((results, data_time), api_result.timestamp, api_result.expires)

    @api.auto_call('map/FacWarSystems')
    def faction_warfare_systems(self, api_result=None):
        """Get a dict of factional warfare systems and their info."""
        rowset = api_result.result.find('rowset')
        results = {}
        for row in rowset.findall('row'):
            system = int(row.attrib['solarSystemID'])
            name = row.attrib['solarSystemName']
            faction_id = int(row.attrib['occupyingFactionID']) or None
            faction_name = row.attrib['occupyingFactionName'] or None
            contested = (row.attrib['contested'] == 'True')

            results[system] = {
                'id': system,
                'name': name,
                'faction': {
                    'id': faction_id,
                    'name': faction_name,
                },
                'contested': contested,
            }

        return api.APIResult(results, api_result.timestamp, api_result.expires)

    @api.auto_call('map/Sovereignty')
    def sov_by_system(self, api_result=None):
        """Get sovereignty info keyed by system."""
        rowset = api_result.result.find('rowset')
        results = {}
        for row in rowset.findall('row'):
            system = int(row.attrib['solarSystemID'])
            name = row.attrib['solarSystemName']
            faction_id = int(row.attrib['factionID']) or None
            alliance_id = int(row.attrib['allianceID']) or None
            corp_id = int(row.attrib['corporationID']) or None

            results[system] = {
                'id': system,
                'name': name,
                'faction_id': faction_id,
                'alliance_id': alliance_id,
                'corp_id': corp_id,
            }

        data_time = api.parse_ts(api_result.result.find('dataTime').text)

        return api.APIResult((results, data_time), api_result.timestamp, api_result.expires)



import collections

from evelink import api, constants
from evelink.parsing.assets import parse_assets
from evelink.parsing.bookmarks import parse_bookmarks
from evelink.parsing.contact_list import parse_contact_list
from evelink.parsing.contract_bids import parse_contract_bids
from evelink.parsing.contract_items import parse_contract_items
from evelink.parsing.contracts import parse_contracts
from evelink.parsing.industry_jobs import parse_industry_jobs
from evelink.parsing.planetary_interactions import parse_planetary_colonies
from evelink.parsing.planetary_interactions import parse_planetary_links
from evelink.parsing.planetary_interactions import parse_planetary_pins
from evelink.parsing.planetary_interactions import parse_planetary_routes
from evelink.parsing.kills import parse_kills
from evelink.parsing.orders import parse_market_orders
from evelink.parsing.wallet_journal import parse_wallet_journal
from evelink.parsing.wallet_transactions import parse_wallet_transactions


class auto_call(api.auto_call):
    """Extends 'evelink.api.auto_call' to add 'Char.char_id' as an api
    request argument.
    """

    def __init__(self, path, map_params=None, **kw):
        map_params = map_params if map_params else {}
        map_params['char_id'] = 'characterID'

        super(auto_call, self).__init__(
            path, prop_to_param=('char_id',), map_params=map_params, **kw
        )


class Char(object):
    """Wrapper around /char/ of the EVE API.

    Note that a valid API key is required.
    """

    def __init__(self, char_id, api):
        self.api = api
        self.char_id = char_id

    @auto_call('char/AssetList')
    def assets(self, api_result=None):
        """Get information about corp assets.

        Each item is a dict, with keys 'id', 'item_type_id',
        'quantity', 'location_id', 'location_flag', and 'packaged'.
        'location_flag' denotes additional information about the
        item's location; see
        http://wiki.eve-id.net/API_Inventory_Flags for more details.

        If the item corresponds to a container, it will have a key
        'contents', which is itself a list of items in the same format
        (potentially recursively holding containers of its own).  If
        the contents do not have 'location_id's of their own, they
        inherit the 'location_id' of their parent container, for
        convenience.

        At the top level, the result is a dict mapping location ID
        (typically a solar system) to a dict containing a 'contents'
        key, which maps to a list of items.  That is, you can think of
        the top-level values as "containers" with no fields except for
        "contents" and "location_id".
        """

        return api.APIResult(parse_assets(api_result.result), api_result.timestamp, api_result.expires)

    @auto_call('char/Bookmarks')
    def bookmarks(self, api_result=None):
        """Retrieves this character's bookmarks."""
        return api.APIResult(parse_bookmarks(api_result.result), api_result.timestamp, api_result.expires)

    @auto_call('char/ContractBids')
    def contract_bids(self, api_result=None):
        """Lists the latest bids that have been made to any recent auctions."""
        return api.APIResult(parse_contract_bids(api_result.result), api_result.timestamp, api_result.expires)

    @auto_call('char/ContractItems', map_params={'contract_id': 'contractID'})
    def contract_items(self, contract_id, api_result=None):
        """Lists items that a specified contract contains"""
        return api.APIResult(parse_contract_items(api_result.result), api_result.timestamp, api_result.expires)

    @auto_call('char/Contracts')
    def contracts(self, api_result=None):
        """Returns a record of all contracts for a specified character"""
        return api.APIResult(parse_contracts(api_result.result), api_result.timestamp, api_result.expires)

    @auto_call('char/WalletJournal', map_params={'before_id': 'fromID', 'limit': 'rowCount'})
    def wallet_journal(self, before_id=None, limit=None, api_result=None):
        """Returns a complete record of all wallet activity for a specified character"""
        return api.APIResult(parse_wallet_journal(api_result.result), api_result.timestamp, api_result.expires)

    @auto_call('char/AccountBalance')
    def wallet_info(self, api_result=None):
        """Return a given character's wallet."""
        rowset = api_result.result.find('rowset')
        row = rowset.find('row')
        result = {
            'balance': float(row.attrib['balance']),
            'id': int(row.attrib['accountID']),
            'key': int(row.attrib['accountKey']),
        }
        return api.APIResult(result, api_result.timestamp, api_result.expires)

    def wallet_balance(self):
        """Helper to return just the balance from a given character wallet"""
        api_result = self.wallet_info()
        return api.APIResult(api_result.result['balance'], api_result.timestamp, api_result.expires)

    @auto_call('char/WalletTransactions', map_params={'before_id': 'fromID', 'limit': 'rowCount'})
    def wallet_transactions(self, before_id=None, limit=None, api_result=None):
        """Returns wallet transactions for a character."""
        return api.APIResult(parse_wallet_transactions(api_result.result), api_result.timestamp, api_result.expires)

    @auto_call('char/IndustryJobs')
    def industry_jobs(self, api_result=None):
        """Get a list of jobs for a character (active only)."""
        return api.APIResult(parse_industry_jobs(api_result.result), api_result.timestamp, api_result.expires)

    @auto_call('char/IndustryJobsHistory')
    def industry_jobs_history(self, api_result=None):
        """Get a historical list of industry jobs for a character (active and past)."""
        return api.APIResult(parse_industry_jobs(api_result.result), api_result.timestamp, api_result.expires)

    @auto_call('char/PlanetaryColonies')
    def planetary_colonies(self, api_result=None):
        """Get a list of PI planets for a character."""
        return api.APIResult(parse_planetary_colonies(api_result.result), api_result.timestamp, api_result.expires)

    @auto_call('char/PlanetaryLinks', map_params={'planet_id': 'planetID'})
    def planetary_links(self, planet_id, api_result=None):
        """Get a list of PI links for a character's planet."""
        return api.APIResult(parse_planetary_links(api_result.result), api_result.timestamp, api_result.expires)

    @auto_call('char/PlanetaryPins', map_params={'planet_id': 'planetID'})
    def planetary_pins(self, planet_id, api_result=None):
        """Get a list of PI facilities for a character's planet."""
        return api.APIResult(parse_planetary_pins(api_result.result), api_result.timestamp, api_result.expires)

    @auto_call('char/PlanetaryRoutes', map_params={'planet_id': 'planetID'})
    def planetary_routes(self, planet_id, api_result=None):
        """Get a list of PI routing entries for a character's planet."""
        return api.APIResult(parse_planetary_routes(api_result.result), api_result.timestamp, api_result.expires)

    def planetary_route_map(self, routes, unused_ts=None, unused_exp=None):
        """Given the result of planetary_routes, build a map planetid: [linkid1, linkid2, ...]"""
        result = collections.defaultdict(set)
        for route_id, route in routes.items():
            result[route['source_id']].add(route_id)
            result[route['destination_id']].add(route_id)

        return dict(result)

    @auto_call('char/KillMails', map_params={'before_kill': 'beforeKillID'})
    def kills(self, before_kill=None, api_result=None):
        """Look up recent kills for a character.

        before_kill:
            Optional. Only show kills before this kill id. (Used for paging.)
        """

        return api.APIResult(parse_kills(api_result.result), api_result.timestamp, api_result.expires)

    @auto_call('char/KillLog', map_params={'before_kill': 'beforeKillID'})
    def kill_log(self, before_kill=None, api_result=None):
        """Look up recent kills for a character.

        Note: this method uses the long cache version of the endpoint. If you
              want to use the short cache version (recommended), use kills().

        before_kill:
            Optional. Only show kills before this kill id. (Used for paging.)
        """

        return api.APIResult(parse_kills(api_result.result), api_result.timestamp, api_result.expires)

    @auto_call('char/Notifications')
    def notifications(self, api_result=None):
        """Returns the message headers for notifications."""
        result = {}
        rowset = api_result.result.find('rowset')
        for row in rowset.findall('row'):
            a = row.attrib
            notification_id = int(a['notificationID'])
            result[notification_id] = {
                'id': notification_id,
                'type_id': int(a['typeID']),
                'sender_id': int(a['senderID']),
                'timestamp': api.parse_ts(a['sentDate']),
                'read': a['read'] == '1',
            }

        return api.APIResult(result, api_result.timestamp, api_result.expires)

    @auto_call('char/NotificationTexts', map_params={'notification_ids': 'IDs'})
    def notification_texts(self, notification_ids, api_result=None):
        """Returns the message bodies for notifications."""
        result = {}
        rowset = api_result.result.find('rowset')
        for row in rowset.findall('row'):
            notification_id = int(row.attrib['notificationID'])
            notification = {'id': notification_id}
            notification.update(api.parse_keyval_data(row.text))
            result[notification_id] = notification

        missing_ids = api_result.result.find('missingIDs')
        if missing_ids is not None:
            for missing_id in missing_ids.text.split(","):
                result[missing_id] = None

        return api.APIResult(result, api_result.timestamp, api_result.expires)

    @auto_call('char/Standings')
    def standings(self, api_result=None):
        """Returns the standings towards a character from NPC entities."""
        result = {}
        rowsets = {}
        for rowset in api_result.result.find('characterNPCStandings').findall('rowset'):
            rowsets[rowset.attrib['name']] = rowset

        _name_map = {
            'agents': 'agents',
            'corps': 'NPCCorporations',
            'factions': 'factions',
        }

        for key, rowset_name in _name_map.items():
            result[key] = {}
            for row in rowsets[rowset_name].findall('row'):
                a = row.attrib
                from_id = int(a['fromID'])
                result[key][from_id] = {
                    'id': from_id,
                    'name': a['fromName'],
                    'standing': float(a['standing']),
                }

        return api.APIResult(result, api_result.timestamp, api_result.expires)

    @auto_call('char/CharacterSheet')
    def character_sheet(self, api_result=None):
        """Returns attributes relating to a specific character."""
        _str, _int, _float, _bool, _ts = api.elem_getters(api_result.result)
        result = {
            'id': _int('characterID'),
            'name': _str('name'),
            'create_ts': _ts('DoB'),
            'race': _str('race'),
            'bloodline': _str('bloodLine'),
            'ancestry': _str('ancestry'),
            'gender': _str('gender'),
            'corp': {
                'id': _int('corporationID'),
                'name': _str('corporationName'),
            },
            'alliance': {
                'id': _int('allianceID') or None,
                'name': _str('allianceName'),
            },
            'balance': _float('balance'),
            'attributes': {},
            'implants': {},
            'jump': {
                'activation_ts': _ts('jumpActivation'),
                'fatigue_ts': _ts('jumpFatigue'),
                'last_update_ts': _ts('jumpLastUpdate'),
            },
            'remote_station_ts': _ts('remoteStationDate'),
            'last_respec_ts': _ts('lastRespecDate'),
            'last_timed_respec_ts': _ts('lastTimedRespec'),
            'free_respecs': _int('freeRespecs'),
            'free_skillpoints': _int('freeSkillPoints'),
            'home_station_id': _int('homeStationID'),
            'jumpclone': {
                'jump_ts': _ts('cloneJumpDate'),
            },
        }

        for attr in ('intelligence', 'memory', 'charisma', 'perception', 'willpower'):
            result['attributes'][attr] = {}
            base = int(api_result.result.findtext('attributes/%s' % attr))
            result['attributes'][attr]['base'] = base

            # NOTE: Removed due to the deprecation of the attribute enhancers section.
            # Better to break things which rely on this field than return the base.
            #
            #result['attributes'][attr]['total'] = base

            # NOTE: CCP has deprecated this in favor of listing the implant typeIDs
            #       as seen below in the 'implants' section.
            #
            #bonus = api_result.result.find('attributeEnhancers/%sBonus' % attr)
            #if bonus is not None:
            #    mod = int(bonus.findtext('augmentatorValue'))
            #    result['attributes'][attr]['total'] += mod
            #    result['attributes'][attr]['bonus'] = {
            #        'name': bonus.findtext('augmentatorName'),
            #        'value': mod,
            #    }

        rowsets = {}
        for rowset in api_result.result.findall('rowset'):
            key = rowset.attrib['name']
            rowsets[key] = rowset

        for implant in rowsets['implants']:
            a = implant.attrib
            result['implants'][int(a['typeID'])] = a['typeName']

        jumpclone_implants = {}
        for implant in rowsets['jumpCloneImplants']:
            a = implant.attrib
            jumpclone_id = int(a['jumpCloneID'])
            implants = jumpclone_implants.setdefault(jumpclone_id, {})
            implants[int(a['typeID'])] = a['typeName']

        result['jumpclone']['clones'] = {}
        for jumpclone in rowsets['jumpClones']:
            a = jumpclone.attrib
            jumpclone_id = int(a['jumpCloneID'])
            location_id = int(a['locationID'])
            # This is keyed off location_id because it simplifies a
            # common lookup ("what systems do I have jumpclones in")
            result['jumpclone']['clones'][location_id] = {
                'id': jumpclone_id,
                'name': a['cloneName'],
                'type_id': int(a['typeID']),
                'location_id': location_id,
                'implants': jumpclone_implants.get(jumpclone_id, {})
            }


        result['skills'] = {}
        result['skillpoints'] = 0
        for skill in rowsets['skills']:
            a = skill.attrib
            skill_id = int(a['typeID'])
            sp = int(a['skillpoints'])
            result['skills'][skill_id] = {
                'id': skill_id,
                'skillpoints': sp,
                'level': int(a['level']),
                'published': a['published'] == '1',
            }
            result['skillpoints'] += sp

        result['roles'] = {}
        for our_role, ccp_role in constants.Char().corp_roles.items():
            result['roles'][our_role] = {}
            for role in rowsets[ccp_role]:
                a = role.attrib
                role_id = int(a['roleID'])
                result['roles'][our_role][role_id] = {
                    'id': role_id,
                    'name': a['roleName'],
                }

        result['titles'] = {}
        for title in rowsets['corporationTitles']:
            a = title.attrib
            title_id = int(a['titleID'])
            result['titles'][title_id] = {
                'id': title_id,
                'name': a['titleName'],
            }

        return api.APIResult(result, api_result.timestamp, api_result.expires)

    @auto_call('char/ContactList')
    def contacts(self, api_result=None):
        """Return a character's personal, corp and alliance contact lists."""
        return api.APIResult(parse_contact_list(api_result.result), api_result.timestamp, api_result.expires)

    @auto_call('char/MarketOrders')
    def orders(self, api_result=None):
        """Return a given character's buy and sell orders."""
        return api.APIResult(parse_market_orders(api_result.result), api_result.timestamp, api_result.expires)

    @auto_call('char/Research')
    def research(self, api_result=None):
        """Returns information about the agents with whom the character is doing research."""
        rowset = api_result.result.find('rowset')
        rows = rowset.findall('row')
        result = {}
        for row in rows:
            a = row.attrib
            id = int(a['agentID'])
            result[id] = {
                'id': id,
                'skill_id': int(a['skillTypeID']),
                'timestamp': api.parse_ts(a['researchStartDate']),
                'per_day': float(a['pointsPerDay']),
                'remaining': float(a['remainderPoints']),
            }

        return api.APIResult(result, api_result.timestamp, api_result.expires)

    @auto_call('char/SkillInTraining')
    def current_training(self, api_result=None):
        """Returns the skill that is currently being trained by a specified character"""
        _str, _int, _float, _bool, _ts = api.elem_getters(api_result.result)
        result = {
            'start_ts': _ts('trainingStartTime'),
            'end_ts': _ts('trainingEndTime'),
            'type_id': _int('trainingTypeID'),
            'start_sp': _int('trainingStartSP'),
            'end_sp': _int('trainingDestinationSP'),
            'current_ts': _ts('currentTQTime'),
            'level': _int('trainingToLevel'),
            'active': _bool('skillInTraining'),
        }

        return api.APIResult(result, api_result.timestamp, api_result.expires)

    @auto_call('char/SkillQueue')
    def skill_queue(self, api_result=None):
        """returns the skill queue of the character"""
        rowset = api_result.result.find('rowset')
        rows = rowset.findall('row')
        result = []
        for row in rows:
            a = row.attrib
            line = {
                'position': int(a['queuePosition']),
                'type_id': int(a['typeID']),
                'level': int(a['level']),
                'start_sp': int(a['startSP']),
                'end_sp': int(a['endSP']),
                'start_ts': api.parse_ts(a['startTime']),
                'end_ts': api.parse_ts(a['endTime']),
            }

            result.append(line)

        return api.APIResult(result, api_result.timestamp, api_result.expires)

    @auto_call('char/MailMessages')
    def messages(self, api_result=None):
        """Returns a list of headers for a character's mail."""
        rowset = api_result.result.find('rowset')
        results = []
        for row in rowset.findall('row'):
            a = row.attrib
            message = {
                'id': int(a['messageID']),
                'sender_id': int(a['senderID']),
                'timestamp': api.parse_ts(a['sentDate']),
                'title': a['title'],
                'to': {},
            }

            org_id = a['toCorpOrAllianceID']
            message['to']['org_id'] = int(org_id) if org_id else None

            char_ids = a['toCharacterIDs']
            message['to']['char_ids'] = [int(i) for i in char_ids.split(',')] if char_ids else None

            list_ids = a['toListID']
            message['to']['list_ids'] = [int(i) for i in list_ids.split(',')] if list_ids else None

            results.append(message)

        return api.APIResult(results, api_result.timestamp, api_result.expires)

    @auto_call('char/MailBodies', map_params={'message_ids': 'ids'})
    def message_bodies(self, message_ids, api_result=None):
        """Returns the actual body content of a set of mail messages.

        NOTE: You *must* have recently looked up the headers of
        any messages you are requesting bodies for (via the 'messages'
        method) or else this call will fail.
        """

        rowset = api_result.result.find('rowset')
        results = {}
        for row in rowset.findall('row'):
            message_id = int(row.attrib['messageID'])
            results[message_id] = row.text

        missing_set = api_result.result.find('missingMessageIDs')
        if missing_set is not None:
            missing_ids = [int(i) for i in missing_set.text.split(',')]
            for missing_id in missing_ids:
                results[missing_id] = None

        return api.APIResult(results, api_result.timestamp, api_result.expires)

    @auto_call('char/MailingLists')
    def mailing_lists(self, api_result=None):
        """Returns the mailing lists to which a character is subscribed."""
        rowset = api_result.result.find('rowset')
        results = {}
        for row in rowset.findall('row'):
            a = row.attrib
            results[int(a['listID'])] = a['displayName']

        return api.APIResult(results, api_result.timestamp, api_result.expires)

    @auto_call('char/UpcomingCalendarEvents')
    def calendar_events(self, api_result=None):
        """Returns the list of upcoming calendar events for a character."""
        results = {}
        rowset = api_result.result.find('rowset')
        for row in rowset.findall('row'):
            a = row.attrib
            event = {
                'id': int(a['eventID']),
                'owner': {
                    'id': int(a['ownerID']),
                    'name': a['ownerName'] or None,
                },
                'start_ts': api.parse_ts(a['eventDate']),
                'title': a['eventTitle'],
                'duration': int(a['duration']),
                'important': a['importance'] == '1',
                'description': a['eventText'],
                'response': a['response'],
            }
            results[event['id']] = event

        return api.APIResult(results, api_result.timestamp, api_result.expires)

    @auto_call('char/CalendarEventAttendees', map_params={'event_ids': 'eventIDs'})
    def calendar_attendees(self, event_ids, api_result=None):
        """Returns the list of attendees for the specified calendar event.

        This function takes a list of event IDs and returns a dict of dicts,
        with the top-level dict being keyed by event ID and the children
        keyed by the character IDs of the attendees.

        NOTE: You must have recently fetched the list of calendar events
        (using the 'calendar_events' method) before calling this method.
        """

        results = dict((int(i),{}) for i in event_ids)
        rowset = api_result.result.find('rowset')
        for row in rowset.findall('row'):
            a = row.attrib
            attendee = {
                'id': int(a['characterID']),
                'name': a['characterName'],
                'response': a['response'],
            }
            results[int(a['eventID'])][attendee['id']] = attendee

        return api.APIResult(results, api_result.timestamp, api_result.expires)

    def event_attendees(self, event_id, api_result=None):
        """Returns the attendees for a single event.

        (This is a convenience wrapper around 'calendar_attendees'.)

        NOTE: You must have recently fetched the list of calendar events
        (using the 'calendar_events' method) before calling this method.
        """

        api_result = self.calendar_attendees([event_id])
        return api.APIResult(api_result.result[int(event_id)], api_result.timestamp, api_result.expires)

    @auto_call('char/FacWarStats')
    def faction_warfare_stats(self, api_result=None):
        """Returns FW stats for this character, if enrolled in FW.

        NOTE: This will return an error instead if the character
        is not enrolled in Faction Warfare.

        """
        _str, _int, _float, _bool, _ts = api.elem_getters(api_result.result)

        result = {
            'faction': {
                'id': _int('factionID'),
                'name': _str('factionName'),
            },
            'enlist_ts': _ts('enlisted'),
            'rank': {
                'current': _int('currentRank'),
                'highest': _int('highestRank'),
            },
            'kills': {
                'yesterday': _int('killsYesterday'),
                'week': _int('killsLastWeek'),
                'total': _int('killsTotal'),
            },
            'points': {
                'yesterday': _int('victoryPointsYesterday'),
                'week': _int('victoryPointsLastWeek'),
                'total': _int('victoryPointsTotal'),
            },
        }

        return api.APIResult(result, api_result.timestamp, api_result.expires)

    @auto_call('char/Medals')
    def medals(self, api_result=None):
        """Returns a list of medals the character has."""
        result = {'current': {}, 'other': {}}
        _map = {
            'currentCorporation': 'current',
            'otherCorporations': 'other',
        }

        for rowset in api_result.result.findall('rowset'):
            name = _map[rowset.attrib['name']]
            for row in rowset.findall('row'):
                a = row.attrib
                medal_id = int(a['medalID'])
                result[name][medal_id] = {
                    'id': medal_id,
                    'reason': a['reason'],
                    'public': a['status'] == 'public',
                    'issuer_id': int(a['issuerID']),
                    'corp_id': int(a['corporationID']),
                    'title': a['title'],
                    'description': a['description'],
                }

        return api.APIResult(result, api_result.timestamp, api_result.expires)

    @auto_call('char/ContactNotifications')
    def contact_notifications(self, api_result=None):
        """Returns pending contact notifications."""
        results = {}
        rowset = api_result.result.find('rowset')
        for row in rowset.findall('row'):
            a = row.attrib
            note = {
                'id': int(a['notificationID']),
                'sender': {
                    'id': int(a['senderID']),
                    'name': a['senderName'],
                },
                'timestamp': api.parse_ts(a['sentDate']),
                'data': api.parse_keyval_data(a['messageData']),
            }
            results[note['id']] = note

        return api.APIResult(results, api_result.timestamp, api_result.expires)

    @auto_call('char/Locations', map_params={'location_list': 'IDs'})
    def locations(self, location_list, api_result=None):
        rowset = api_result.result.find('rowset')
        rows = rowset.findall('row')

        results = {}
        for row in rows:
            name = row.attrib['itemName'] or None
            id = int(row.attrib['itemID']) or None
            x = float(row.attrib['x']) or None
            y = float(row.attrib['y']) or None
            z = float(row.attrib['z']) or None

            results[id] = {
                'name': name,
                'id' : id,
                'x' : x,
                'y' : y,
                'z' : z,
            }
        return api.APIResult(results, api_result.timestamp, api_result.expires)

    @auto_call('char/Blueprints')
    def blueprints(self, api_result=None):
        rowset = api_result.result.find('rowset')
        rows = rowset.findall('row')

        results = {}
        for row in rows:
            results[int(row.attrib['itemID'])] = {
                'location_id': int(row.attrib['locationID']),
                'type_id': int(row.attrib['typeID']),
                'type_name': row.attrib['typeName'],
                'location_flag': int(row.attrib['flagID']),
                'quantity': int(row.attrib['quantity']),
                'time_efficiency': int(row.attrib['timeEfficiency']),
                'material_efficiency': int(row.attrib['materialEfficiency']),
                'runs': int(row.attrib['runs']),
            }

        return api.APIResult(results, api_result.timestamp, api_result.expires)

    @auto_call('char/ChatChannels')
    def chat_channels(self, api_result=None):
        """Get a list of chat channels this character owns or ops."""
        rowset = api_result.result.find('rowset')

        results = {}
        for row in rowset.findall('row'):
            a = row.attrib
            channel = {
                'id': int(a['channelID']),
                'owner': {
                    'id': int(a['ownerID']),
                    'name': a['ownerName'],
                },
                'name': a['displayName'],
                'comparison_name': a['comparisonKey'],
                'passworded': a['hasPassword'] == 'True',
                'motd': a['motd'],
                'allowed': {},
                'blocked': {},
                'muted': {},
                'ops': {},
            }

            sections = {}
            for section in row.findall('rowset'):
                sections[section.attrib['name']] = section

            if 'allowed' in sections:
                for entity in sections['allowed'].findall('row'):
                    entity_id = int(entity.attrib['accessorID'])
                    channel['allowed'][entity_id] = {
                        'id': entity_id,
                        'name': entity.attrib['accessorName'],
                    }
            if 'blocked' in sections:
                for entity in sections['blocked'].findall('row'):
                    entity_id = int(entity.attrib['accessorID'])
                    channel['blocked'][entity_id] = {
                        'id': entity_id,
                        'name': entity.attrib['accessorName'],
                        'until_ts': api.parse_ts(entity.attrib['untilWhen']),
                        'reason': entity.attrib['reason'],
                    }
            if 'muted' in sections:
                for entity in sections['muted'].findall('row'):
                    entity_id = int(entity.attrib['accessorID'])
                    channel['muted'][entity_id] = {
                        'id': entity_id,
                        'name': entity.attrib['accessorName'],
                        'until_ts': api.parse_ts(entity.attrib['untilWhen']),
                        'reason': entity.attrib['reason'],
                    }
            if 'operators' in sections:
                for entity in sections['operators'].findall('row'):
                    entity_id = int(entity.attrib['accessorID'])
                    channel['ops'][entity_id] = {
                        'id': entity_id,
                        'name': entity.attrib['accessorName'],
                    }

            results[channel['id']] = channel

        return api.APIResult(results, api_result.timestamp, api_result.expires)


# vim: set ts=4 sts=4 sw=4 et:



from evelink import api

def parse_bookmarks(api_results):
    result = {}
    folders = api_results.find('rowset')
    for row in folders.findall('row'):
        a = row.attrib
        folder_id = int(a['folderID'])
        folder = {
            'id': folder_id,
            'name': a['folderName'], # "" = toplevel
            'bookmarks': {},
        }
        bookmarks = row.find('rowset')
        for row in bookmarks.findall('row'):
            a = row.attrib
            bookmark_id = int(a['bookmarkID'])
            folder['bookmarks'][bookmark_id] = {
                'id': bookmark_id,
                'name': a['memo'],
                'creator_id': int(a['creatorID']),
                'created_ts': api.parse_ts(a['created']),
                'item_id': int(a['itemID']),
                'type_id': int(a['typeID']),
                'location_id': int(a['locationID']),
                'x': float(a['x']),
                'y': float(a['y']),
                'z': float(a['z']),
                'note': a['note'],
            }
        result[folder_id] = folder
    return result

# vim: set ts=4 sts=4 sw=4 et:



import calendar
import collections
import functools
import zlib
import inspect
import logging
import re
import time
import hashlib
from xml.etree import ElementTree

from evelink.thirdparty import six
from evelink.thirdparty.six.moves import urllib

_log = logging.getLogger('evelink.api')

# Python 2.6's ElementTree raises xml.parsers.expat.ExpatError instead
# of ElementTree.ParseError
_xml_error = getattr(ElementTree, 'ParseError', None)
if _xml_error is None:
    import xml.parsers.expat
    _xml_error = xml.parsers.expat.ExpatError

# Allows zlib.decompress to decompress gzip-compressed strings as well.
# From zlib.h header file, not documented in Python.
ZLIB_DECODE_AUTO = 32 + zlib.MAX_WBITS

# Set by evelink/__init__.py to the evelink version. Use the user_agent
# parameter when constructing an API object if you want to add additional
# information to the user agent string. (Technically, you *can* override
# this, but it's not the intended method.)
_user_agent = None

# Can be set to an APICache instance that is used as a shared default
# cache instance for all API instances. Note: instance, not class.
default_cache = None

# The timeout to use for API HTTP requests, in seconds (default 1 minute).
http_request_timeout = 60

try:
    import requests
    _has_requests = True
except ImportError:
    _log.info('`requests` not available, falling back to urllib2')
    _has_requests = None

def _clean(v):
    """Convert parameters into an acceptable format for the API."""
    if isinstance(v, (list, set, tuple)):
        return ",".join(str(i) for i in v)
    else:
        return str(v)

def decompress(s):
    """Decode a gzip compressed string."""
    return zlib.decompress(s, ZLIB_DECODE_AUTO)


def parse_ts(v):
    """Parse a timestamp from EVE API XML into a unix-ish timestamp."""
    if v == '':
        return None
    ts = calendar.timegm(time.strptime(v, "%Y-%m-%d %H:%M:%S"))
    # Deal with EVE's nonexistent 0001-01-01 00:00:00 timestamp
    return ts if ts > 0 else None


def get_named_value(elem, field):
    """Returns the string value of the named child element."""
    try:
        return elem.find(field).text
    except AttributeError:
        return None


def get_ts_value(elem, field):
    """Returns the timestamp value of the named child element."""
    val = get_named_value(elem, field)
    if val:
        return parse_ts(val)
    return None


def get_int_value(elem, field):
    """Returns the integer value of the named child element."""
    val = get_named_value(elem, field)
    if val:
        return int(val)
    return val


def get_float_value(elem, field):
    """Returns the float value of the named child element."""
    val = get_named_value(elem, field)
    if val:
        return float(val)
    return val


def get_bool_value(elem, field):
    """Returns the boolean value of the named child element."""
    val = get_named_value(elem, field)
    if val == 'True':
        return True
    elif val == 'False':
        return False
    return None


def elem_getters(elem):
    """Returns a tuple of (_str, _int, _float, _bool, _ts) functions.

    These are getters closed around the provided element.
    """
    _str = lambda key: get_named_value(elem, key)
    _int = lambda key: get_int_value(elem, key)
    _float = lambda key: get_float_value(elem, key)
    _bool = lambda key: get_bool_value(elem, key)
    _ts = lambda key: get_ts_value(elem, key)

    return _str, _int, _float, _bool, _ts


def parse_keyval_data(data_string):
    """Parse 'key: value' lines from a LF-delimited string."""
    keyval_pairs = data_string.strip().split('\n')
    results = {}
    for pair in keyval_pairs:
        key, _, val = pair.strip().partition(': ')

        if 'Date' in key:
            val = parse_ms_date(val)
        elif val == 'null':
            val = None
        elif re.match(r"^-?\d+$", val):
            val = int(val)
        elif re.match(r"-?\d+\.\d+", val):
            val = float(val)

        results[key] = val
    return results

def parse_ms_date(date_string):
    """Convert MS date format into epoch"""

    return int(date_string)/10000000 - 11644473600;

class APIError(Exception):
    """Exception raised when the EVE API returns an error."""

    def __init__(self, code=None, message=None, timestamp=None, expires=None):
        self.code = code
        self.message = message
        self.timestamp = timestamp
        self.expires = expires

    def __repr__(self):
        return "APIError(%r, %r, timestamp=%r, expires=%r)" % (
            self.code, self.message, self.timestamp, self.expires)

    def __str__(self):
        return "%s (code=%d)" % (self.message, int(self.code))

class APICache(object):
    """Minimal interface for caching API requests.

    This very basic implementation simply stores values in
    memory, with no other persistence. You can subclass it
    to define a more complex/featureful/persistent cache.
    """

    def __init__(self):
        self.cache = {}

    def get(self, key):
        """Return the value referred to by 'key' if it is cached.

        key:
            a string hash key
        """
        result = self.cache.get(key)
        if not result:
            return None
        value, expiration = result
        if expiration < time.time():
            del self.cache[key]
            return None
        return value

    def put(self, key, value, duration):
        """Cache the provided value, referenced by 'key', for the given duration.

        key:
            a string hash key
        value:
            an xml.etree.ElementTree.Element object
        duration:
            a number of seconds before this cache entry should expire.
        """
        expiration = time.time() + duration
        self.cache[key] = (value, expiration)


APIResult = collections.namedtuple("APIResult", [
        "result",
        "timestamp",
        "expires",
    ])


class API(object):
    """A wrapper around the EVE API."""

    def __init__(self, base_url="api.eveonline.com", cache=None, api_key=None, user_agent=None):
        self.base_url = base_url
        self.user_agent = _user_agent

        if user_agent is not None:
            self.user_agent += ' %s' % user_agent

        cache = cache or default_cache or APICache()
        if not isinstance(cache, APICache):
            raise ValueError("The provided cache must subclass from APICache.")
        self.cache = cache
        self.CACHE_VERSION = '1'

        if api_key and len(api_key) != 2:
            raise ValueError("The provided API key must be a tuple of (keyID, vCode).")
        self.api_key = api_key
        self._set_last_timestamps()

    def _set_last_timestamps(self, current_time=0, cached_until=0):
        self.last_timestamps = {
            'current_time': current_time,
            'cached_until': cached_until,
        }

    def _cache_key(self, path, params):
        sorted_params = sorted(params.items())
        # Paradoxically, Shelve doesn't like integer keys.
        return '%s-%s' % (self.CACHE_VERSION, hashlib.sha1(str([path,sorted_params]).encode("utf-8")).hexdigest())

    def get(self, path, params=None):
        """Request a specific path from the EVE API.

        The supplied path should be a slash-separated path
        frament, e.g. "corp/AssetList". (Basically, the portion
        of the API url in between the root / and the .xml bit.)
        """

        params = params or {}
        params = dict((k, _clean(v)) for k,v in params.items())

        _log.debug("Calling %s with params=%r", path, params)
        if self.api_key:
            _log.debug("keyID and vCode added")
            params['keyID'] = self.api_key[0]
            params['vCode'] = self.api_key[1]

        key = self._cache_key(path, params)
        response = self.cache.get(key)
        cached = response is not None

        if not cached:
            # no cached response body found, call the API for one.
            full_path = "https://%s/%s.xml.aspx" % (self.base_url, path)
            response, robj = self.send_request(full_path, params)
        else:
            _log.debug("Cache hit, returning cached payload")

        try:
            tree = ElementTree.fromstring(response)
        except _xml_error as e:
            # If this is due to an HTTP error, raise the HTTP error
            self.maybe_raise_http_error(robj)
            # otherwise, raise the parse error
            raise e

        current_time = get_ts_value(tree, 'currentTime')
        expires_time = get_ts_value(tree, 'cachedUntil')
        self._set_last_timestamps(current_time, expires_time)

        if not cached:
            # Have to split this up from above as timestamps have to be
            # extracted.
            self.cache.put(key, response, expires_time - current_time)

        error = tree.find('error')
        if error is not None:
            code = error.attrib['code']
            message = error.text.strip()
            exc = APIError(code, message, current_time, expires_time)
            _log.debug("Raising API error: %r" % exc)
            raise exc

        result = tree.find('result')
        return APIResult(result, current_time, expires_time)

    def maybe_raise_http_error(self, response):
        """Called if a XML parse error is raised for the response.

        Raises an error if the response itself was an error - we try
        to parse it as XML first to see if it's an API error, but if
        it's not, this method gets called.
        """
        if _has_requests:
            # Requests has a built-in method for this functionality
            response.raise_for_status()
        else:
            # urllib2 uses exceptions by default for this, which we've
            # potentially previously caught and stored as the response
            if isinstance(response, Exception):
                raise response

    def send_request(self, full_path, params):
        if _has_requests:
            return self.requests_request(full_path, params)
        else:
            return self.urllib2_request(full_path, params)

    def urllib2_request(self, full_path, params):
        r = None
        try:
            if params:
                # POST request
                _log.debug("POSTing request")
                params = urllib.parse.urlencode(params)
                req = urllib.request.Request(full_path, data=params.encode())
            else:
                # GET request
                req = urllib.request.Request(full_path)
                _log.debug("GETting request")

            req.add_header('Accept-Encoding', 'gzip')
            req.add_header('User-agent', self.user_agent)
            r = urllib.request.urlopen(req, timeout=http_request_timeout)
        except urllib.error.HTTPError as e:
            # urllib2 handles non-2xx responses by raising an exception that
            # can also behave as a file-like object. The EVE API will return
            # non-2xx HTTP codes on API errors (since Odyssey, apparently)
            r = e
        except urllib.error.URLError as e:
            # TODO: Handle this better?
            raise e

        try:
            if r.info().get('Content-Encoding') == 'gzip':
                return decompress(r.read()), r
            else:
                return r.read(), r
        finally:
            r.close()

    def requests_request(self, full_path, params):
        session = getattr(self, 'session', None)
        if not session:
            session = requests.Session()
            session.headers.update({'User-Agent': self.user_agent})
            self.session = session

        try:
            if params:
                # POST request
                _log.debug("POSTing request")
                r = session.post(full_path, data=params, timeout=http_request_timeout)
            else:
                # GET request
                _log.debug("GETting request")
                r = session.get(full_path, timeout=http_request_timeout)
            _log.debug("Response status code: %s" % r.status_code)
            return r.content, r
        except requests.exceptions.RequestException as e:
            # TODO: Handle this better?
            raise e


def auto_api(func):
    """A decorator to automatically provide an API instance.

    Functions decorated with this will have the api= kwarg
    automatically supplied with a default-initialized API()
    object if no other API object is supplied.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        defaultargs, defaultkwargs = get_args_and_defaults(func)
        mapped_args = map_func_args(args, kwargs, defaultargs, defaultkwargs)
        if mapped_args.get('api') is None:
            kwargs['api'] = API()
        return func(*args, **kwargs)
    return wrapper


def translate_args(args, mapping=None):
    """Translate python name variable into API parameter name."""
    mapping = mapping if mapping else {}
    return dict((mapping[k], v,) for k, v in args.items())

# TODO: needs better name
def get_args_and_defaults(func):
    """Return the list of argument names and a dict of default values"""
    specs = inspect.getargspec(func)
    return (
        specs.args,
        dict(zip(specs.args[-len(specs.defaults):], specs.defaults)),
    )


def map_func_args(args, kw, args_names, defaults):
    """Associate positional (*args) and key (**kw) arguments values
    with their argument names.

    'args_names' should be the list of argument names and 'default'
    should be a dict associating the keyword arguments to their
    defaults.

    Similar to inspect.getcallargs() but compatible with python 2.6.

    """
    if (len(args)+len(kw)) > len(args_names):
        raise TypeError('Too many arguments.')

    map_ = dict(zip(args_names, args))
    for k, v in kw.items():
        if k in map_:
            raise TypeError(
                "got multiple values for keyword argument '%s'" % k
            )
        map_[k] = v

    for k, v in defaults.items():
        map_.setdefault(k, v)

    required_args = args_names[0:-len(defaults)]
    for k in required_args:
        if k not in map_:
            raise TypeError("Too few arguments")
    return map_


class auto_call(object):
    """A decorator to automatically provide an api response to a method.

    The object the method will be bound to should have an api attribute
    and the method should have a keyword argument named 'api_result'.

    The decorated method will have a '_request_specs' dict attribute
    holding:

    - 'path': path of the request needs to be queried.

    - 'args': method argument names.

    - 'defaults': method keyword arguments and theirs default value.

    - 'prop_to_param': properties of the instance the method is bound
    to to add as parameter of api request.

    - 'map_params': dictionary associating argument name to a
    paramater name. They will be added to 'evelink.api._args_map' to
    translate argument names to parameter names.

    """

    def __init__(self, path, prop_to_param=tuple(), map_params=None):
        self.method = None

        self.path = path
        self.args = None
        self.defaults = None
        self.prop_to_param = prop_to_param
        self.map_params = map_params if map_params else {}

    def __call__(self, method):
        if self.method is not None:
            raise TypeError("This decorator method cannot be shared.")
        self.method = method

        wrapper = self._wrapped_method()

        args, self.defaults = get_args_and_defaults(self.method)

        self.args = args[1:]
        self.args.remove('api_result')
        self.defaults.pop('api_result')  # TODO: better exception

        wrapper._request_specs = {
            'path': self.path,
            'args': self.args,
            'defaults': self.defaults,
            'prop_to_param': self.prop_to_param,
            'map_params': self.map_params
        }

        return wrapper

    def _wrapped_method(self):

        @functools.wraps(self.method)
        def wrapper(client, *args, **kw):
            if 'api_result' in kw:
                return self.method(client, *args, **kw)

            args_map = map_func_args(args, kw, self.args, self.defaults)
            for attr_name in self.prop_to_param:
                args_map[attr_name] = getattr(client, attr_name, None)

            params = translate_args(args_map, self.map_params)
            params =  dict((k, v,) for k, v in params.items() if v is not None)

            kw['api_result'] = client.api.get(self.path, params=params)
            return self.method(client, *args, **kw)

        return wrapper


# vim: set ts=4 sts=4 sw=4 et:



from google.appengine.ext import ndb

from evelink import eve, api
from evelink.appengine.api import auto_async, auto_gae_api

@auto_async
class EVE(eve.EVE):
    __doc__ = eve.EVE.__doc__

    @auto_gae_api
    def __init__(self, api=None):
        self.api = api

    @ndb.tasklet
    def character_name_from_id_async(self, char_id):
        """Asynchronous version of character_name_from_id."""
        resp = yield self.character_names_from_ids_async([char_id])
        raise ndb.Return(
            api.APIResult(
                resp.result.get(char_id), resp.timestamp, resp.expires
            )
        )

    @ndb.tasklet
    def character_id_from_name_async(self, name):
        """Asynchronous version of character_id_from_name."""
        resp = yield self.character_ids_from_names_async([name])
        raise ndb.Return(
            api.APIResult(resp.result.get(name), resp.timestamp, resp.expires)
        )



from evelink import map as map_
from evelink.appengine.api import auto_async, auto_gae_api


@auto_async
class Map(map_.Map):
    __doc__ = map_.Map.__doc__

    @auto_gae_api
    def __init__(self, api=None):
        self.api = api



from google.appengine.ext import ndb

from evelink import corp
from evelink.appengine.api import auto_async


@auto_async
class Corp(corp.Corp):
    __doc__ = corp.Corp.__doc__

    @ndb.tasklet
    def members_async(self, extended=True):
        """Returns details about each member of the corporation."""
        args = {}
        if extended:
            args['extended'] = 1
        
        api_result = yield self.api.get_async(
        	'corp/MemberTracking', params=args
        )
        raise ndb.Return(
        	self.members(extended=extended, api_result=api_result)
        )


from evelink import api

def parse_planetary_colonies(api_results):
    result = {}
    rowset = api_results.find('rowset')
    for row in rowset.findall('row'):
        # shortcut to make the following block less painful
        a = row.attrib
        planetID = int(a['planetID'])
        result[planetID] = {
            'id': planetID,
            'system': {
                'id': int(a['solarSystemID']),
                'name': a['solarSystemName'],
            },
            'planet': {
                'name': a['planetName'],
                'type': int(a['planetTypeID']),
                'type_name': a['planetTypeName'],
            },
            'owner': {
                'id': int(a['ownerID']),
                'name': a['ownerName'],
            },
            'last_update': api.parse_ts(a['lastUpdate']),
            'upgrade_level': int(a['upgradeLevel']),
            'number_of_pins': int(a['numberOfPins']),
        }

    return result


def parse_planetary_links(api_results):
    result = {}
    rowset = api_results.find('rowset')
    for row in rowset.findall('row'):
        # shortcut to make the following block less painful
        a = row.attrib
        sourceID = int(a['sourcePinID'])
        result[sourceID] = {
            'source_id': sourceID,
            'destination_id': int(a['destinationPinID']),
            'link_level': int(a['linkLevel']),
        }

    return result

def parse_planetary_pins(api_results):
    result = {}
    rowset = api_results.find('rowset')
    for row in rowset.findall('row'):
        # shortcut to make the following block less painful
        a = row.attrib
        pinID = int(a['pinID'])
        if pinID not in result:
            result[pinID] = {
                'id': pinID,
                'type': {
                    'id': int(a['typeID']),
                    'name': a['typeName'],
                },
                'schematic': int(a['schematicID']),
                'last_launch_ts': api.parse_ts(a['lastLaunchTime']),
                'cycle_time': int(a['cycleTime']),
                'quantity_per_cycle': int(a['quantityPerCycle']),
                'install_ts': api.parse_ts(a['installTime']),
                'expiry_ts': api.parse_ts(a['expiryTime']),
                'content': {
                    'type': int(a['contentTypeID']),
                    'name': a['contentTypeName'],
                    'quantity': int(a['contentQuantity']),
                    'deprecated': 'Use the "contents" field instead',
                },
                'contents': {},
                'loc': {
                    'long': float(a['longitude']),
                    'lat': float(a['latitude']),
                },
            }

        if a['contentTypeID'] != '0':
            contentID = int(a['contentTypeID'])
            result[pinID]['contents'][contentID] = {
                'type': contentID,
                'name': a['contentTypeName'],
                'quantity': int(a['contentQuantity']),
            }


    return result

def parse_planetary_routes(api_results):
    result = {}
    rowset = api_results.find('rowset')
    for row in rowset.findall('row'):
        # shortcut to make the following block less painful
        a = row.attrib
        routeID = int(a['routeID'])
        result[routeID] = {
            'id': routeID,
            'source_id': int(a['sourcePinID']),
            'destination_id': int(a['destinationPinID']),
            'content': {
                'type': int(a['contentTypeID']),
                'name': a['contentTypeName'],
            },
            'quantity': int(a['quantity']),
            'path': tuple(int(a['waypoint%d' % n]) for n in range(1,6)),
        }

    return result



#!/usr/bin/python

import os
from setuptools import setup

from evelink import __version__

try:
    __readme_path = os.path.join(os.path.dirname(__file__), "README.md")
    __readme_contents = open(__readme_path).read()
except IOError:
    __readme_contents = ""

setup(
    name="EVELink",
    version=__version__,
    description="Python Bindings for the EVE Online API",
    long_description=__readme_contents,
    license="MIT License",
    author="Valkyries of Night",
    author_email="d-eve-lopment@googlegroups.com",
    maintainer="Amber Yust",
    maintainer_email="amber.yust@gmail.com",
    url="https://github.com/eve-val/evelink",
    download_url="https://github.com/eve-val/evelink/downloads",
    packages=[
        "evelink",
        "evelink.appengine",
        "evelink.cache",
        "evelink.parsing",
        "evelink.thirdparty",
    ],
    data_files=[
        ('', ['README.md', 'LICENSE']),
    ],
    scripts=["bin/evelink"],
    provides=["evelink"],
    requires=["six"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Topic :: Games/Entertainment",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)

# vim: set et ts=4 sts=4 sw=4:



from evelink import server
from evelink.appengine.api import auto_async, auto_gae_api

@auto_async
class Server(server.Server):
    __doc__ = server.Server.__doc__

    @auto_gae_api
    def __init__(self, api=None):
        self.api = api



from google.appengine.ext import ndb

from evelink import char, api
from evelink.appengine.api import auto_async

@auto_async
class Char(char.Char):
    __doc__ = char.Char.__doc__

    @ndb.tasklet
    def wallet_balance_async(self):
        """Asynchronous version of wallet_balance."""
        api_result = yield self.wallet_info_async()
        raise ndb.Return(
            api.APIResult(
                api_result.result['balance'],
                api_result.timestamp,
                api_result.expires
            )
        )

    @ndb.tasklet
    def event_attendees_async(self, event_id, api_result=None):
        """Asynchronous version of event_attendees."""
        api_result = yield self.calendar_attendees_async([event_id])
        raise ndb.Return(
            api.APIResult(
                api_result.result[int(event_id)],
                api_result.timestamp,
                api_result.expires
            )
        )
